/* 
    jgaadb - is a simple C++ class library for ODBC databases.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/

#include "StdAfx.h"
#include "jgaadb.h"
#include <assert.h>
#include <time.h>


BEGIN_JGAADB_NAMESPACE

/////////////////////////////// PUBLIC ///////////////////////////////////////

#define DEFAULT_LARGE_BUFFER_LEN (1024 * 512) // 512 Kbytes

//============================= LIFECYCLE ====================================

DbColumn::DbColumn(const char *name,
                   const int colSqlType, 
                   const size_t colMaxSize,
                   const size_t colDigits,
                   const bool isNullable,
                   const DbDataTypesE colType) :
mSqlType(colSqlType),
mSize(colMaxSize),
mByteSize(0),
mDigits(colDigits),
mIsNullable(isNullable),
mAllocatedBytes(0),
mType(colType),
mName(name ? name : ""),
mDataLenUsed(0),
mUnknownBufferSize(false),
mCharWidth(0),
mIsStringType(false)
{
    memset(&mData, 0, sizeof(mData));

    mCompatibleName = get_compatible_name(mName.c_str());

    if (DB_TYPE_INVALID == mType)
    {
        switch(mSqlType)
        {
        case SQL_CHAR: 
            mType = DB_TYPE_CHAR;
            mByteSize = (colMaxSize + 1 ) * sizeof(db_char_t);
            mCharWidth = sizeof(db_char_t);
            mIsStringType = true;
            break;
        case SQL_VARCHAR: 
            mType = DB_TYPE_VARCHAR;
            mByteSize = (colMaxSize + 1 ) * sizeof(db_char_t);
            mCharWidth = sizeof(db_char_t);
            mIsStringType = true;
            break;
        case SQL_LONGVARCHAR: 
            mType = DB_TYPE_TEXT;
            if ((mByteSize = (colMaxSize + 1 ) * sizeof(db_char_t)) > DEFAULT_LARGE_BUFFER_LEN)
                mByteSize = DEFAULT_LARGE_BUFFER_LEN; 
            mUnknownBufferSize = true;
            mCharWidth = sizeof(db_char_t);
            mIsStringType = true;
            break;
        case SQL_WCHAR: 
            mType = DB_TYPE_WCHAR;
            mByteSize = (colMaxSize + 1 ) * sizeof(db_wchar_t);
            mCharWidth = sizeof(db_wchar_t);
            mIsStringType = true;
            break;
        case SQL_WVARCHAR: 
            mType = DB_TYPE_VARWCHAR;
            mByteSize = (colMaxSize + 1 ) * sizeof(db_wchar_t);
            mCharWidth = sizeof(db_wchar_t);
            mIsStringType = true;
            break;
        case SQL_WLONGVARCHAR: 
            mType = DB_TYPE_WTEXT;
            if ((mByteSize = (colMaxSize + 1 ) * sizeof(db_wchar_t)) > DEFAULT_LARGE_BUFFER_LEN)
                mByteSize = DEFAULT_LARGE_BUFFER_LEN;
            mUnknownBufferSize = true;
            mCharWidth = sizeof(db_wchar_t);
            mIsStringType = true;
            break;
        case SQL_DECIMAL: 
            mType = DB_TYPE_DECIMAL;
            mByteSize = 128;
            break;
        case SQL_NUMERIC: 
            mType = DB_TYPE_NUMERIC;
            mByteSize = 128;
            break;
        case SQL_SMALLINT: 
            mType = DB_TYPE_SHORT;
            break;
        case SQL_INTEGER: 
            mType = DB_TYPE_INT;
            break;
        case SQL_REAL:
            mType = DB_TYPE_FLOAT;
            break;
        case SQL_FLOAT: 
            mType = DB_TYPE_FLOAT;
            break;
        case SQL_DOUBLE: 
            mType = DB_TYPE_DOUBLE;
            break;
        case SQL_BIT: 
            mType = DB_TYPE_INT;
            break;
        case SQL_TINYINT: 
            mType = DB_TYPE_INT;
            break;
        case SQL_BIGINT: 
            mType = DB_TYPE_BIGINT;
            break;
        case SQL_BINARY: 
            mType = DB_TYPE_BIN;
            mByteSize = colMaxSize;
            break;
        case SQL_VARBINARY: 
            mType = DB_TYPE_VARBIN;
            mByteSize = colMaxSize;
            break;
        case SQL_LONGVARBINARY: 
            mType = DB_TYPE_BLOB;
            mByteSize = colMaxSize;
            break;
		case SQL_DATE:
        case SQL_TYPE_DATE: 
            mType = DB_TYPE_DATE;
			if (0 == mSize)
				mSize = 10;
            break;
		case SQL_TIME:
        case SQL_TYPE_TIME: 
            mType = DB_TYPE_TIME;
			if (0 == mSize)
				mSize = 8;
            break;
        case SQL_TYPE_TIMESTAMP: 
		case SQL_TIMESTAMP:
            mType = DB_TYPE_TIMESTAMP;
			if (0 == mSize)
				mSize = 19;
            break;
        default:
            mType = DB_TYPE_VARCHAR;
            break;
        }
    }

    if (colMaxSize)
	{
        Allocate(mByteSize);
	}

}// DbColumn


DbColumn::~DbColumn()
{
   Clear();
}

//============================= OPERATORS ====================================

void DbColumn::operator = (const db_int_t& val)
{
    if (DB_TYPE_INT != mType)
    {
		std::string msg = "Not DB_TYPE_INT in column ";
		msg += GetName();
		DbThrow(msg.c_str());
    }

    mData.db_int = val;
    mDataLenUsed = sizeof(db_int_t);
}

void DbColumn::operator = (const db_short_t& val)
{
    if (DB_TYPE_SHORT != mType)
    {
		std::string msg = "Not DB_TYPE_SHORT in column ";
		msg += GetName();
		DbThrow(msg.c_str());
    }

    mData.db_short = val;
    mDataLenUsed = sizeof(db_short_t);
}

void DbColumn::operator = (const db_tinyint_t& val)
{
    if (DB_TYPE_TINYINT != mType)
    {
		std::string msg = "Not DB_TYPE_TINYINT in column ";
		msg += GetName();
		DbThrow(msg.c_str());
    }

    mData.db_tinyint = val;
    mDataLenUsed = sizeof(db_tinyint_t);
}


void DbColumn::operator = (const db_i64_t& val)
{
    if (DB_TYPE_BIGINT == mType)
	{
		mData.db_i64 = val;
		mDataLenUsed = sizeof(db_i64_t);
	}
	else if ((DB_TYPE_BIN == mType)
		|| (DB_TYPE_VARBIN == mType))
	{
		// Check length. Must be 8 bytes
		if (GetSize() != 8)
		{
			DbThrow("DB_TYPE_BIN = db_i64_t, but length is not 8 bytes!");
		}

		db_i64_t *p = (db_i64_t *)Allocate(8);
		*p = val;
		mDataLenUsed = 8;
	}
	else
    {
		std::string msg = "Not DB_TYPE_INT or DB_TYPE_BIN in column ";
		msg += GetName();
		DbThrow(msg.c_str());
    }
}

void DbColumn::operator = (const time_t& val)
{
	const struct tm *tm = ::gmtime((time_t *)&val);
	static const struct tm tm_buf;
	if (!tm)
		tm = &tm_buf;

	if (DB_TYPE_TIME == mType)
	{
		mData.db_time.hour = tm->tm_hour;
		mData.db_time.minute = tm->tm_min;
		mData.db_time.second = tm->tm_sec;

		mDataLenUsed = sizeof(db_time_t);
	}
	else if (DB_TYPE_DATE == mType)
	{	 
		mData.db_date.year = tm->tm_year + 1900;
		mData.db_date.month = tm->tm_mon + 1;
		mData.db_date.day = tm->tm_mday;

		mDataLenUsed = sizeof(db_date_t);
	}
	else if (DB_TYPE_TIMESTAMP == mType)
	{
		mData.db_timestamp.hour = tm->tm_hour;
		mData.db_timestamp.minute = tm->tm_min;
		mData.db_timestamp.second = tm->tm_sec;
		mData.db_timestamp.year = tm->tm_year + 1900;
		mData.db_timestamp.month = tm->tm_mon + 1;
		mData.db_timestamp.day = tm->tm_mday;
		mData.db_timestamp.fraction = 0;

		mDataLenUsed = sizeof(db_timestamp_t);
	}
	else
	{
		std::string msg = "Not DB_TYPE_TIME or DB_TYPE_DATE or DB_TYPE_TIMESTAMP in column ";
		msg += GetName();
		DbThrow(msg.c_str());
	}
}

void DbColumn::operator = (const char * val)
{
    if ((DB_TYPE_VARWCHAR == mType)
		|| (DB_TYPE_WCHAR == mType))
    {
        size_t len = strlen(val);
        
        std::vector<wchar_t> buffer(len + 16);
        
        int bytes = (int)::mbstowcs(&buffer[0], val, len);
        operator = ((const wchar_t *)&buffer[0]);
        return;
    }
	else if (DB_TYPE_CHAR == mType)
	{
		size_t len = strlen(val),
			cp_len = len;
		if (cp_len > mSize)
			cp_len = mSize;
		char *pbuf = (char *)Allocate(mSize +1);
		strncpy(pbuf, val, cp_len);
		// Add padding
		memset(pbuf + cp_len, ' ', mSize - cp_len);
		pbuf[mSize] = 0;
		mDataLenUsed = mSize;
	}
	else if (DB_TYPE_VARCHAR == mType)
	{
		size_t len = strlen(val) + 1;
		memcpy(Allocate(len), val, len);
		mDataLenUsed = len;
		if (mSize < len)
			mSize = len;
	}
	else
    {
		std::string msg = "Not DB_TYPE_*CHAR in column ";
		msg += GetName();
		DbThrow(msg.c_str());
    }
}

void DbColumn::operator = (const wchar_t * val)
{
    if (DB_TYPE_VARWCHAR != mType)
    {
		std::string msg = "Not DB_TYPE_VARWCHAR in column ";
		msg += GetName();
		DbThrow(msg.c_str());
    }

    size_t len = 0;
    const wchar_t *p = val;
    while(*p++)
        len++;
    len++;
    len *= sizeof(wchar_t);

    memcpy(Allocate(len), val, len);
    mDataLenUsed = len;
    
    if (mSize < len)
        mSize = len;
}

//============================= OPERATIONS ===================================

void DbColumn::ResetCache()
{
    if (!mCachedString.empty())
        mCachedString.erase();
}

std::string DbColumn::GetCreateStatement(DbConnection *pTarget) const
{
    std::stringstream query;
    std::string data_type_name;

    query << GetCompatibleName() << ' ';
    
    if (pTarget)
    {
        // Check if the destination have the datatype
        DbCommand my_cmd(*pTarget);
        DbRecordset my_dataset(&my_cmd);
        if (!JGAADB_SQLOK(SQLGetTypeInfo(my_cmd.GetHtsmt(), mSqlType)))
        {
            // Not supported
            switch(mSqlType)
            {
            case SQL_WCHAR:
                goto do_char;
            case SQL_WVARCHAR:
                goto do_varchar;
            case SQL_WLONGVARCHAR:
                goto do_text;
            }
            goto default_handle;
        }
        my_dataset.FetchFirst();
        try
        {
            data_type_name = my_dataset[0].GetAsString();
        }
        catch(std::exception)
        {
            data_type_name = "_ ILLEGAL NAME";
        }
    }
    
    switch(mSqlType)
    {
    case SQL_CHAR: 
do_char:
        query << "CHAR(" << mSize << ") ";
        break;
    case SQL_VARCHAR: 
do_varchar:
        query << "VARCHAR(" << mSize << ") ";
        break;
    case SQL_LONGVARCHAR: 
do_text:
        query << "TEXT ";
        break;
#ifdef SQL_WCHAR
    case SQL_WCHAR: 
        if (!data_type_name.empty() && stricmp(data_type_name.c_str(), "WCHAR"))
            goto do_char;
        else
            query << "WCHAR(" << mSize << ") ";
        break;
#endif
#ifdef SQL_WVARCHAR
    case SQL_WVARCHAR: 
        if (!data_type_name.empty() && stricmp(data_type_name.c_str(), "WVARCHAR"))
            goto do_varchar;
        else
            query << "WVARCHAR(" << mSize << ") ";
        break;
#endif
#ifdef SQL_WLONGVARCHAR
    case SQL_WLONGVARCHAR: 
        if (!data_type_name.empty() && stricmp(data_type_name.c_str(), "WTEXT"))
            goto do_text;
        else
            query << "WTEXT ";
        break;
#endif
    case SQL_DECIMAL: 
        query << "DECIMAL(" << mSize << ',' << mDigits << ") ";
        break;
    case SQL_NUMERIC: 
        query << "NUMERIC(" << mSize << ',' << mDigits << ") ";
        break;
    case SQL_SMALLINT: 
        query << "SMALLINT(" << mSize << ") ";
        break;
    case SQL_INTEGER: 
        query << "INTEGER(" << mSize << ") ";
        break;
    case SQL_REAL:
        query << "REAL ";
        break;
    case SQL_FLOAT: 
        query << "FLOAT ";
        break;
    case SQL_DOUBLE: 
        query << "FLOAT ";
        break;
    case SQL_BIT: 
        query << "TINYINT ";
        break;
    case SQL_TINYINT: 
        query << "TINYINT ";
        break;
    case SQL_BIGINT: 
        query << "BIGINT ";
        break;
    case SQL_BINARY: 
        query << "BINARY(" << mSize << ") ";
        break;
    case SQL_VARBINARY: 
        query << "BINARY(" << mSize << ") ";
        break;
    case SQL_LONGVARBINARY: 
        query << "BLOB ";
        break;
    case SQL_TYPE_DATE: 
        query << "DATE ";
        break;
    case SQL_TYPE_TIME: 
        query << "TIME ";
        break;
    case SQL_TYPE_TIMESTAMP: 
        query << "DATETIME ";
        break;
    default:
default_handle:
        query << "VARCHAR(255) ";
        break;
    }

    if (!mIsNullable)
        query << "NOT NULL ";

    return query.str();
}

//============================= CALLBACK   ===================================
//============================= ACCESS     ===================================

void * DbColumn::GetPointer() throw(DbException)
{
    void *rval = NULL;

    switch(mType)
    {
    case DB_TYPE_TINYINT:
        rval = &mData.db_tinyint;
        break;
    case DB_TYPE_UTINYINT:
        rval = &mData.db_utinyint;
        break;
    case DB_TYPE_BIGINT:
        rval = &mData.db_i64;
        break;
    case DB_TYPE_UBIGINT:
        rval = &mData.db_u64;
        break;
    case DB_TYPE_SHORT:
        rval = &mData.db_short;
        break;
    case DB_TYPE_INT:
        rval = &mData.db_int;
        break;
    case DB_TYPE_USHORT:
        rval = &mData.db_ushort;
        break;
    case DB_TYPE_UINT:
        rval = &mData.db_uint;
        break;
    case DB_TYPE_CHAR: 
    case DB_TYPE_VARCHAR:
    case DB_TYPE_TEXT:
    case DB_TYPE_BLOB:
    case DB_TYPE_BIN:
    case DB_TYPE_VARBIN:
    case DB_TYPE_DECIMAL:
    case DB_TYPE_NUMERIC:
        rval = mData.db_cstr;
        break;
    case DB_TYPE_VARWCHAR:
    case DB_TYPE_WCHAR:
    case DB_TYPE_WTEXT:
        rval = mData.db_wstr;
        break;    
    case DB_TYPE_TIME:
        rval = &mData.db_time;
        break;
    case DB_TYPE_DATE:
        rval = &mData.db_date;
        break;
    case DB_TYPE_TIMESTAMP:
        rval = &mData.db_timestamp;
        break;
    case DB_TYPE_FLOAT:
        rval = &mData.db_float;
        break;
    case DB_TYPE_DOUBLE:
    case DB_TYPE_MONEY:
        rval = &mData.db_double;
        break;
    default:
        DbThrow("Invalid type on column");
    }

    return rval;
}

void DbColumn::GetBindBuffer(void **pbuffer, 
                             size_t*buf_len, 
                             int *c_type,
                             int *sql_type,
                             DbConnection *pTarget) 
                             throw(std::exception)
{
    if (pTarget)
    {
        // Check if the destination have the datatype
        DbCommand my_cmd(*pTarget);
        DbRecordset my_dataset(&my_cmd);
        if (!JGAADB_SQLOK(SQLGetTypeInfo(my_cmd.GetHtsmt(), mSqlType)))
        {
            switch(mType)
            {
            case DB_TYPE_WCHAR:
                *sql_type = SQL_CHAR;
                *pbuffer = (void *)GetAsString();
                *buf_len = strlen((char *)*pbuffer);
                *c_type = SQL_C_CHAR;
                return;
            case DB_TYPE_VARWCHAR:
                 *sql_type = SQL_VARCHAR;
                *pbuffer = (void *)GetAsString();
                *buf_len = strlen((char *)*pbuffer);
                *c_type = SQL_C_CHAR;
                return;
            case DB_TYPE_WTEXT:
                 *sql_type = SQL_LONGVARCHAR;
                *pbuffer = (void *)GetAsString();
                *buf_len = strlen((char *)*pbuffer);
                *c_type = SQL_C_CHAR;
                return;
            default:
                break;
            }
        }
    }

    *buf_len = mDataLenUsed;
    *c_type = GetCtype();
    *sql_type = mSqlType;

    switch(mType)
    {
    case DB_TYPE_TINYINT:
        *pbuffer = &mData.db_tinyint;
        break;
    case DB_TYPE_UTINYINT:
        *pbuffer = &mData.db_utinyint;
        break;
    case DB_TYPE_BIGINT:
        *pbuffer = &mData.db_i64;
        break;
    case DB_TYPE_UBIGINT:
        *pbuffer = &mData.db_u64;
        break;
    case DB_TYPE_SHORT:
        *pbuffer = &mData.db_short;
        break;
    case DB_TYPE_INT:
        *pbuffer = &mData.db_int;
        break;
    case DB_TYPE_USHORT:
        *pbuffer = &mData.db_ushort;
        break;
    case DB_TYPE_UINT:
        *pbuffer = &mData.db_uint;
        break;
    case DB_TYPE_CHAR:
        *sql_type = SQL_CHAR;
    case DB_TYPE_VARCHAR:
    case DB_TYPE_TEXT:
    case DB_TYPE_DECIMAL:
    case DB_TYPE_NUMERIC:
        *pbuffer = (void *)GetAsString();
        *buf_len = strlen((char *)*pbuffer);
        *c_type = SQL_C_CHAR;
        break;
    case DB_TYPE_BIN:
		if (0 == *buf_len)
			*buf_len = mSize;
		// Fall trough
    case DB_TYPE_VARBIN: // Can't be used in prepared statements
	case DB_TYPE_BLOB: // Can't be used in prepared statements
        *pbuffer = mData.db_cstr;
        break;
    case DB_TYPE_VARWCHAR:
    case DB_TYPE_WCHAR:
    case DB_TYPE_WTEXT:
        *pbuffer = mData.db_wstr;
        break;    
    case DB_TYPE_TIME:
        *pbuffer = &mData.db_time;
        break;
    case DB_TYPE_DATE:
        *pbuffer = &mData.db_date;
        break;
    case DB_TYPE_TIMESTAMP:
        *pbuffer = &mData.db_timestamp;
        break;
    case DB_TYPE_FLOAT:
        *pbuffer = &mData.db_float;
        break;
    case DB_TYPE_DOUBLE:
    case DB_TYPE_MONEY:
        *pbuffer = &mData.db_double;
        break;
    default:
        DbThrow("Invalid type on column");
    }
}

const int DbColumn::GetAsInt() const throw(std::exception)
{
	if (DB_TYPE_INT == GetType())
	{
		return mData.db_int;
	}
	else if (DB_TYPE_TINYINT == mType)
	{
		return mData.db_tinyint;
	}
	else
	{
		std::string msg = "Not DB_TYPE_*INT in column ";
		msg += GetName();
		DbThrow(msg.c_str());
	}

	return 0; // Compiler food
}

const char *DbColumn::GetAsString() throw(std::exception)
{
    if (!mCachedString.empty())
        return mCachedString.c_str();

    char buffer[128];
    buffer[0] = 0;

    if (SQL_NULL_DATA != mDataLenUsed) switch(mType)
    {
    case DB_TYPE_TINYINT:
        sprintf(buffer, "%i", (int)mData.db_tinyint);
        break;
    case DB_TYPE_UTINYINT:
        sprintf(buffer, "%u", (unsigned)mData.db_utinyint);
        break;
    case DB_TYPE_BIGINT:
        sprintf(buffer, "%i64", mData.db_i64);
        break;
    case DB_TYPE_UBIGINT:
        sprintf(buffer, "%u64",mData.db_u64);
        break;
    case DB_TYPE_SHORT:
        sprintf(buffer, "%i", (int)mData.db_short);
        break;
    case DB_TYPE_INT:
        sprintf(buffer, "%i", mData.db_int);
        break;
    case DB_TYPE_USHORT:
        sprintf(buffer, "%u", (unsigned)mData.db_ushort);
        break;
    case DB_TYPE_UINT:
        sprintf(buffer, "%u", mData.db_uint);
        break;
    case DB_TYPE_CHAR: 
    case DB_TYPE_VARCHAR:
    case DB_TYPE_TEXT:
    case DB_TYPE_DECIMAL:
    case DB_TYPE_NUMERIC:
        return mData.db_cstr;
    case DB_TYPE_BLOB:
    case DB_TYPE_BIN:
    case DB_TYPE_VARBIN:
        {
            // Hex encode
            std::vector<char> buffer((mDataLenUsed * 2) + 1);
            char *p = &buffer[0];
            const char *pp = mData.db_cstr;
            for(size_t index = 0; index < mDataLenUsed; index++)
            {
                sprintf(p, "%02x", ((unsigned)*pp++) & 0xff);
                p += 2;
            }
            *p = 0;
            mCachedString = (const char *)&buffer[0];

#ifdef _DEBUG
            int buf_len = buffer.size(), str_len = mCachedString.size();
            assert(buf_len == (str_len +1));
#endif
            
            return mCachedString.c_str();
        }
    case DB_TYPE_VARWCHAR:
    case DB_TYPE_WCHAR:
    case DB_TYPE_WTEXT:
        {
            std::vector<char> buffer(mDataLenUsed + 16);
            int bytes = (int)::wcstombs(&buffer[0], mData.db_wstr, buffer.size());
            if (bytes > 0)
                mCachedString = (const char *)&buffer[0];

            return mCachedString.c_str();
        }
        break;    
    case DB_TYPE_TIME:
        sprintf(buffer, "%02u:%02u:%02u", 
            mData.db_time.hour, 
            mData.db_time.minute,
            mData.db_time.second);
        break;
    case DB_TYPE_DATE:
        sprintf(buffer, "%04d-%02d-%02d",
            mData.db_date.year,
            mData.db_date.month,
            mData.db_date.day);
        break;
    case DB_TYPE_TIMESTAMP:
        sprintf(buffer, "%04d-%02d-%02d %02u:%02u:%02u",
            mData.db_timestamp.year,
            mData.db_timestamp.month,
            mData.db_timestamp.day,
            mData.db_timestamp.hour, 
            mData.db_timestamp.minute,
            mData.db_timestamp.second);
        break;
    case DB_TYPE_FLOAT:
        sprintf(buffer, "%f", mData.db_float);
        break;
    case DB_TYPE_DOUBLE:
    case DB_TYPE_MONEY:
        sprintf(buffer, "%lf", mData.db_double);
        break;
    default:
        DbThrow("Invalid type on column");
    }

    mCachedString = buffer;
    return mCachedString.c_str();
}

//============================= INQUIRY    ===================================


SQLSMALLINT DbColumn::GetCtype() const throw(std::exception)
{
    if ((0 > mType)
        || (DB_TYPE_INVALID <= mType))
    {
        DbThrow("DbColumn::GetCtype(): Type is out of range");
    }

    static const int type_map[DB_TYPE_INVALID] = 
    {
        SQL_C_STINYINT, //DB_TYPE_TINYINT,
        SQL_C_UTINYINT, // DB_TYPE_TINYINT,
        SQL_C_SBIGINT, // DB_TYPE_BIGINT,
        SQL_C_UBIGINT, // DB_TYPE_UBIGINT,
        SQL_C_SSHORT, //DB_TYPE_SHORT,
        SQL_C_SLONG, // DB_TYPE_INT,
        SQL_C_USHORT, // DB_TYPE_USHORT,
        SQL_C_ULONG, // DB_TYPE_UINT,
        SQL_C_CHAR, // DB_TYPE_CHAR, 
        SQL_C_CHAR, // DB_TYPE_VARCHAR,
        SQL_C_WCHAR, // DB_TYPE_WCHAR,
        SQL_C_WCHAR, // DB_TYPE_VARWCHAR,
        SQL_C_TYPE_DATE, //DB_TYPE_DATE,
        SQL_C_TYPE_TIME, // DB_TYPE_TIME,
        SQL_C_TYPE_TIMESTAMP, //DB_TYPE_TIMESTAMP,
        SQL_C_BINARY, // DB_TYPE_BIN,
        SQL_C_BINARY, // DB_TYPE_VARBIN
        SQL_C_FLOAT, // DB_TYPE_FLOAT,
        SQL_C_DOUBLE, // DB_TYPE_DOUBLE,
        SQL_C_DOUBLE, // DB_TYPE_MONEY,
        SQL_C_CHAR, // DB_TYPE_TEXT,
        SQL_C_WCHAR, // DB_TYPE_WTEXT,
        SQL_C_BINARY, // DB_TYPE_BLOB,
        SQL_C_CHAR, // DB_TYPE_DECIMAL,
        SQL_C_CHAR // DB_TYPE_NUMERIC
    };

    return type_map[mType];
}

size_t DbColumn::GetBufferLength() const
{
    if (mByteSize)
        return mByteSize;

    return sizeof(mData);
}

const char *DbColumn::GetName() const
{
    return mName.c_str();
}

const char *DbColumn::GetCompatibleName() const
{
    return mCompatibleName.c_str();
}

DbDataTypesE DbColumn::GetType() const
{
    return mType;
}

const char *DbColumn::GetTypeName() const
{
    static const char *type_names[] = 
    {
        {"DB_TYPE_TINYINT"},
        {"DB_TYPE_UTINYINT"},
        {"DB_TYPE_BIGINT"},
        {"DB_TYPE_UBIGINT"},
        {"DB_TYPE_SHORT"},
        {"DB_TYPE_INT"},
        {"DB_TYPE_USHORT"},
        {"DB_TYPE_UINT"},
        {"DB_TYPE_CHAR"}, 
        {"DB_TYPE_VARCHAR"},
        {"DB_TYPE_WCHAR"},
        {"DB_TYPE_VARWCHAR"},
        {"DB_TYPE_DATE"},
        {"DB_TYPE_TIME"},
        {"DB_TYPE_TIMESTAMP"},
        {"DB_TYPE_BIN"},
        {"DB_TYPE_VARBIN"},
        {"DB_TYPE_FLOAT"},
        {"DB_TYPE_DOUBLE"},
        {"DB_TYPE_MONEY"},
        {"DB_TYPE_TEXT"},
        {"DB_TYPE_WTEXT"},
        {"DB_TYPE_BLOB"},
        {"DB_TYPE_INVALID"}
    };

     if ((0 > mType)
        || (DB_TYPE_INVALID < mType))
    {
        DbThrow("DbColumn::GetTypeName(): Type is out of range");
    }

    return type_names[mType];
}

size_t DbColumn::GetSize() const
{
    return mSize;
}

bool DbColumn::IsStringType() const
{
    return mIsStringType;
}

size_t DbColumn::GetCharWidth() const
{
    return mCharWidth;
}

bool DbColumn::IsNull() const
{
    return (SQL_NULL_DATA == mDataLenUsed);
}

SQLSMALLINT DbColumn::GetSqlType() const throw(std::exception)
{
    return mSqlType;
}

SQLSMALLINT DbColumn::GetDigits() const
{
    return mDigits;
}

SQLSMALLINT DbColumn::GetBytesUsed() const
{
    return mDataLenUsed;
}

/////////////////////////////// PROTECTED  ///////////////////////////////////

void * DbColumn::Allocate(size_t bytes,
                          bool doCopyOldBuffer) 
                          throw(DbException)
{
    switch(mType)
    {
    case DB_TYPE_CHAR:
    case DB_TYPE_VARCHAR:
    case DB_TYPE_TEXT:
    case DB_TYPE_BIN:
    case DB_TYPE_VARBIN:
    case DB_TYPE_BLOB:
    case DB_TYPE_DECIMAL:
    case DB_TYPE_NUMERIC:
        {
			if (bytes > mAllocatedBytes)
			{
				db_cstr_t p = new db_char_t[bytes];
				*p = 0; // Terminate empty string
				if (mData.db_cstr)
				{
					if (doCopyOldBuffer)
						memcpy(p, mData.db_cstr, DB_MIN(mAllocatedBytes, bytes));
					delete mData.db_cstr;
				}
				mData.db_cstr = p;
				mAllocatedBytes = bytes;
			}
        }
        break;
    case DB_TYPE_WCHAR:
    case DB_TYPE_VARWCHAR:
    case DB_TYPE_WTEXT:
        {
			if (bytes > mAllocatedBytes)
			{
				db_wstr_t p = new db_wchar_t[bytes];
				*p = 0; // Terminate empty string
				if (mData.db_wstr)
				{
					if (doCopyOldBuffer)
						memcpy(p, mData.db_wstr, DB_MIN(mAllocatedBytes, bytes));
					delete mData.db_wstr;
				}
				mData.db_wstr = p;
				mAllocatedBytes = bytes;
			}
        }
        break;
    default:
        memset(&mData, 0, sizeof(mData));
    }

    return GetPointer();
}

void DbColumn::Clear()
{
#define KILL(a) {if (NULL != a) { delete a; a = NULL; }}
    switch(mType)
    {
    case DB_TYPE_CHAR:
    case DB_TYPE_VARCHAR:
    case DB_TYPE_TEXT:
    case DB_TYPE_BIN:
    case DB_TYPE_VARBIN:
    case DB_TYPE_BLOB:
        KILL(mData.db_cstr);
        break;
    case DB_TYPE_WCHAR:
    case DB_TYPE_VARWCHAR:
    case DB_TYPE_WTEXT:
        KILL(mData.db_wstr);
        break;
    default:
        memset(&mData, 0, sizeof(mData));
    }

    mAllocatedBytes = 0;

#undef KILL
}

/////////////////////////////// PRIVATE    ///////////////////////////////////

END_JGAADB_NAMESPACE

