/* 
    jgaadb - is a simple C++ class library for ODBC databases.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/

#include "StdAfx.h"
#include "jgaadb.h"


/////////////////////////////// PUBLIC ///////////////////////////////////////

BEGIN_JGAADB_NAMESPACE

int DbCommand::msDefaultTimeout = 0;

//============================= LIFECYCLE ====================================

DbCommand::DbCommand(DbConnection& conn) :
mIsVirgin(true), mConn(conn), mHstmt(NULL), mAffectedRows(-1), mParamIndex(0),
mTimeout(msDefaultTimeout), mDebugFlag(false)
{
    AllocHstmt();
}// DbCommand

DbCommand::~DbCommand()
{
    Close();
}// ~DbCommand


//============================= OPERATORS ====================================
//============================= OPERATIONS ===================================

void DbCommand::Close()
{
    if (mHstmt)
    {
        SQLFreeStmt(mHstmt, SQL_CLOSE);
#if defined(HAVE_SQLFREEHANDLE) || defined(WIN32)
        SQLFreeHandle(SQL_HANDLE_STMT, mHstmt);
#endif
        mHstmt = NULL;
    }
}

void DbCommand::Reset()
{
    AllocHstmt();
}

void DbCommand::SetTimeout(SQLUINTEGER newTimeout)
{
    mTimeout = newTimeout;
}

void DbCommand::Execute(const char *query, 
                        DbRecordset *pResultset)
                        throw(std::exception)
{
    if (pResultset)
    {
        pResultset->Reset();
        pResultset->mpCommand = this;
    }

    if (!mIsVirgin)
    {
        DbThrow("DbCommand::Execute(): I'm not a virgin anymore!");
    }

    if (!JGAADB_SQLOK(SQLSetStmtAttr(mHstmt,
        SQL_ATTR_QUERY_TIMEOUT,
        &mTimeout,
        SQL_IS_UINTEGER)))
    {
        DbException ex(&mConn,
            SQL_HANDLE_STMT,
            mHstmt,
            "DbCommand::Execute()",
            "SQLSetStmtAttr(SQL_ATTR_QUERY_TIMEOUT) failed.",
            NULL,
			__LINE__, __FILE__);

		// Examine the error
		std::string state;
		ex.GetSqlState(state);

		if (strcmp(state.c_str(), "HYC00") == 0)
		{
			// Optional feature not implemented
			;
		}
		else if (strcmp(state.c_str(), "IM001") == 0)
		{
			// Unsupported by driver
			;
		}
		else
		{
			throw ex;
		}
    }

    if (!JGAADB_SQLOK(SQLExecDirect(
        mHstmt, (unsigned char*)query, SQL_NTS)))
    {
        ThrowDb(&mConn,
            SQL_HANDLE_STMT,
            mHstmt,
            "DbCommand::Execute()",
            "SQLExecDirect() failed.",
            query);
    }

    mIsVirgin = false;
    mAffectedRows = -1;
    SQLRowCount(mHstmt, &mAffectedRows);

    if (pResultset)
        pResultset->FetchFirst();
}

void DbCommand::SetDebugMode(bool enableDebug)
{
    mDebugFlag = enableDebug;
}

void DbCommand::PrepareStatement(const char *query)
throw(std::exception)
{
    if (!mIsVirgin)
    {
        DbThrow("DbCommand::Execute(): I'm not a virgin anymore!");
    }

    if (mDebugFlag)
    {
        SQLINTEGER out_len = 0;
        SQLCHAR native_sql_buf[1024 * 16];
        if (!JGAADB_SQLOK(SQLNativeSql(mConn.GetHdbc(), (SQLCHAR *)query, SQL_NTS,
            native_sql_buf, sizeof(native_sql_buf), &out_len)))
        {
            ThrowDb(&mConn,
                SQL_HANDLE_DBC,
                mConn.GetHdbc(),
                "DbCommand::PrepareStatement()",
                "SQLNativeSql() failed.",
                query);
        }

        std::clog << "DbCommand::PrepareStatement(): " << native_sql_buf;
    }

    if (!JGAADB_SQLOK(SQLPrepare(mHstmt, (SQLCHAR *)query, SQL_NTS)))
    {
        ThrowDb(&mConn,
            SQL_HANDLE_STMT,
            mHstmt,
            "DbCommand::PrepareStatement()",
            "SQLPrepare() failed.",
            query);
    }

    mIsVirgin = false;
}

void DbCommand::ExecutePrepared(DbRecordset *pResultset)
throw(std::exception)
{
    if (pResultset)
    {
        pResultset->Reset();
        pResultset->mpCommand = this;
    }

    if (mIsVirgin)
    {
        DbThrow("DbCommand::Execute(): I'm a virgin!");
    }

    SQLRETURN status = SQLExecute(mHstmt);

    if ((SQL_SUCCESS != status) && (SQL_NO_DATA != status))
    {
        if ((SQL_ERROR == status) || (SQL_SUCCESS_WITH_INFO == status))
        {
            ThrowDb(&mConn,
                SQL_HANDLE_STMT,
                mHstmt,
                "DbCommand::ExecutePrepared()",
                "SQLExecute() failed.",
                NULL);
        }

        if (SQL_NEED_DATA == status)
        {
            ThrowDb(&mConn,
                SQL_HANDLE_STMT,
                mHstmt,
                "DbCommand::ExecutePrepared()",
                "SQLExecute() failed: SQL_NEED_DATA",
                NULL);
        }

        if (SQL_NEED_DATA == status)
        {
            ThrowDb(&mConn,
                SQL_HANDLE_STMT,
                mHstmt,
                "DbCommand::ExecutePrepared()",
                "SQLExecute() failed: SQL_INVALID_HANDLE",
                NULL);
        }
        
        ThrowDb(&mConn,
            SQL_HANDLE_STMT,
            mHstmt,
            "DbCommand::ExecutePrepared()",
            "SQLExecute() failed: Undocumented error",
            NULL);
    }

    mAffectedRows = -1;
    SQLRowCount(mHstmt, &mAffectedRows);

    if (pResultset)
        pResultset->FetchFirst();
}

void DbCommand::ResetArguments()
{
    mParamIndex = 0;
}


void DbCommand::AllocHstmt()
throw(std::exception)
{
    Close();
    mAffectedRows = -1;
    mIsVirgin = true;
    mParamIndex = 0;

#if defined(HAVE_SQLALLOCHANDLE) || defined(WIN32)
    if (!JGAADB_SQLOK(SQLAllocHandle(
        SQL_HANDLE_STMT, mConn.GetHdbc(), &mHstmt)))
#else
    if (!JGAADB_SQLOK(SQLAllocStmt(mConn.GetHdbc(), &mHstmt)))
#endif
    {
        ThrowDb(&mConn,
            SQL_HANDLE_DBC,
            mConn.GetHdbc(),
            "DbCommand::AllocHstmt()",
            "SQLAllocHandle(SQL_HANDLE_STMT) failed.",
            NULL);
    }
}


void DbCommand::EnumTables(db_table_list_t& tables)
throw(DbException)
{
    SQLUINTEGER catalog_name_len = 156,
        schema_name_len = 256, table_name_len = 256,
        type_name_len = 256;
    AllocHstmt();

    if (!JGAADB_SQLOK(SQLTables(mHstmt,
        NULL, 0, NULL, 0, (unsigned char*)"%", SQL_NTS, NULL, 0)))
    {
        ThrowDb(&mConn,
            SQL_HANDLE_STMT,
            mHstmt,
            "DbCommand::EnumTables()",
            "SQLTables() failed.",
            NULL);
    }

#if defined(SQL_MAX_CATALOG_NAME_LEN)
    if (!JGAADB_SQLOK(SQLGetInfo(mConn.GetHdbc(), 
        SQL_MAX_CATALOG_NAME_LEN, 
        (SQLPOINTER)&catalog_name_len, sizeof(catalog_name_len), NULL))
        || !JGAADB_SQLOK(SQLGetInfo(mConn.GetHdbc(), 
        SQL_MAX_SCHEMA_NAME_LEN, 
        (SQLPOINTER)&schema_name_len, sizeof(schema_name_len), NULL))
        || !JGAADB_SQLOK(SQLGetInfo(mConn.GetHdbc(), 
        SQL_MAX_TABLE_NAME_LEN, 
        (SQLPOINTER)&table_name_len, sizeof(table_name_len), NULL)))
    {
        ThrowDb(&mConn,
            SQL_HANDLE_DBC,
            mConn.GetHdbc(),
            "DbCommand::EnumTables()",
            "SQLGetInfo() failed.",
            NULL);
    }
#endif

    std::vector<char>catalog_name(catalog_name_len +1);
    std::vector<char>schema_name(schema_name_len +1);
    std::vector<char>table_name(table_name_len +1);
    std::vector<char>type_name(type_name_len +1);
    SQLINTEGER catalog_cb = 0, schema_cb = 0, table_cb = 0, type_cb = 0;

    if (!JGAADB_SQLOK(SQLBindCol(mHstmt, 1, SQL_C_CHAR, 
        (SQLPOINTER)&catalog_name[0], catalog_name_len, &catalog_cb))
        || !JGAADB_SQLOK(SQLBindCol(mHstmt, 2, SQL_C_CHAR, 
        (SQLPOINTER)&schema_name[0], schema_name_len, &schema_cb))
        || !JGAADB_SQLOK(SQLBindCol(mHstmt, 3, SQL_C_CHAR, 
        (SQLPOINTER)&table_name[0], table_name_len, &table_cb))
        || !JGAADB_SQLOK(SQLBindCol(mHstmt, 4, SQL_C_CHAR, 
        (SQLPOINTER)&type_name[0], type_name_len, &type_cb)))
    {
    }

    for(;;)
    {
        SQLRETURN status = SQLFetch(mHstmt);
        if (!JGAADB_SQLOK(status))
        {
#if SQL_NO_DATA
            if (SQL_NO_DATA == status)
                break;
#endif

#if SQL_NO_DATA_FOUND
            if (SQL_NO_DATA_FOUND == status)
                break;
#endif
            
            ThrowDb(&mConn,
                SQL_HANDLE_STMT,
                mHstmt,
                "DbCommand::EnumTables()",
                "SQLFetch() failed.",
                NULL);
        }

        tables.push_back(DbTable(&table_name[0],
            &schema_name[0],
            &catalog_name[0],
            &type_name[0]));
    }
}

void DbCommand::Bind(DbColumn *pCol) throw(std::exception)
{
    void *ptr = NULL;
    size_t buf_len = 0;
    int c_type = 0, sql_type = 0;

    pCol->GetBindBuffer(&ptr, &buf_len, &c_type, &sql_type, &GetConn());

	if (pCol->IsNull())
	{
		pCol->mDataLenUsed2 = SQL_NULL_DATA;
	}
	else if (SQL_C_CHAR == c_type)
	{
		buf_len = pCol->GetBufferLength();
		pCol->mDataLenUsed2 = SQL_NTS;
	}
	else
		pCol->mDataLenUsed2 = buf_len;

	SQLSMALLINT col_size = pCol->GetSize();
	SQLSMALLINT col_digits = pCol->GetDigits();

    if (!JGAADB_SQLOK(SQLBindParameter(GetHtsmt(), 
        ++mParamIndex,
        SQL_PARAM_INPUT,
        c_type,
        sql_type,
        col_size,
        col_digits,
        ptr,
        buf_len,
        &pCol->mDataLenUsed2)))
    {
        ThrowDb(&mConn,
            SQL_HANDLE_STMT,
            mHstmt,
            "DbCommand::Bind()",
            "SQLBindParameter() failed.",
            NULL);
    }
}

//============================= CALLBACK   ===================================
//============================= ACCESS     ===================================

SQLHSTMT DbCommand::GetHtsmt()
{
    return mHstmt;
}

DbConnection &DbCommand::GetConn()
{
    return mConn;
}

//============================= INQUIRY    ===================================

long DbCommand::GetAffectedRows() const
{
    return mAffectedRows;
}

/////////////////////////////// PROTECTED  ///////////////////////////////////

/////////////////////////////// PRIVATE    ///////////////////////////////////


END_JGAADB_NAMESPACE

