/* 
    jgaadb - is a simple C++ class library for ODBC databases.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/

#include "StdAfx.h"
#include "jgaadb.h"   // class implemented

/////////////////////////////// PUBLIC ///////////////////////////////////////
BEGIN_JGAADB_NAMESPACE

//============================= LIFECYCLE ====================================

DbConnection::DbConnection()
: mHdbc(NULL), mHenv(NULL), mOdbcVersion(ODBC_V2), mIsConnected(false)
{
}// DbConnection


DbConnection::~DbConnection()
{
    Close();
}// ~DbConnection


//============================= OPERATORS ====================================
//============================= OPERATIONS ===================================

void DbConnection::Connect(const char *Dsn, int loginTimeout)
        throw(DbException)
{
    Close();

    mDsn = Dsn;
	SQLRETURN status; 

#if defined(WIN32) || defined(HAVE_SQLALLOCHANDLE)
    status = SQLAllocHandle(SQL_HANDLE_ENV, 
        SQL_NULL_HANDLE, 
        &mHenv);
#else
    status = SQLAllocEnv(&mHenv);
#endif

	if (!JGAADB_SQLOK(status))
	{
        ThrowDb(this, 
            0,
            NULL,
            "DbConnection::Connect()",
            "SQLAllocHandle(SQL_HANDLE_ENV) failed.",
            NULL);
	}

#if defined(SQL_ATTR_ODBC_VERSION)
	status = SQLSetEnvAttr(mHenv,SQL_ATTR_ODBC_VERSION, (SQLPOINTER)SQL_OV_ODBC3, sizeof(SQLPOINTER));
    if (JGAADB_SQLOK(status))
        mOdbcVersion = ODBC_V3;
#endif

#if defined(SQL_LOGIN_TIMEOUT) && (defined(HAVE_SQLSETCONNECTATTR) || defined(WIN32))
	status = SQLSetConnectAttr(mHenv, SQL_LOGIN_TIMEOUT, (SQLPOINTER)loginTimeout, 0);
#endif	
	

#if defined(WIN32) || defined(HAVE_SQLALLOCHANDLE)
    status = SQLAllocHandle(SQL_HANDLE_DBC, mHenv, &mHdbc);
#else
    status = SQLAllocConnect(mHenv, &mHdbc);
#endif
	if (!JGAADB_SQLOK(status))
	{
        ThrowDb(this,
            SQL_HANDLE_ENV,
            GetHenv(),
            "DbConnection::Connect()",
            "SQLAllocHandle(SQL_HANDLE_DBC) failed.",
            NULL);
    }

    SQLSMALLINT out_clen = 0;
    SQLCHAR cbuf[1024];
    status = SQLDriverConnect(mHdbc, 
        NULL, 
        (SQLCHAR *)mDsn.c_str(),
        mDsn.size(), 
        cbuf,
        sizeof(cbuf), 
        &out_clen,
        SQL_DRIVER_NOPROMPT);
    
    if (!JGAADB_SQLOK(status))
    {
        ThrowDb(this,
            SQL_HANDLE_DBC,
            GetHdbc(),
            "DbConnection::Connect()",
            "SQLDriverConnect(SQL_HANDLE_DBC) failed.",
            NULL);
    }

    mConnectDsn = (const char *)cbuf;
    mIsConnected = true;

	// Get some info
	QueryInfo(SQL_DBMS_NAME, mDbmsName);
	QueryInfo(SQL_DBMS_VER, mDbmsVersion);
	QueryInfo(SQL_IDENTIFIER_QUOTE_CHAR, mSqlQuoteChar);
}

void DbConnection::Close()
{
    if (mHdbc)
    {
        if (mIsConnected)
            SQLDisconnect(mHdbc);
#if defined(HAVE_SQLFREEHANDLE) || defined(WIN32)
        SQLFreeHandle(SQL_HANDLE_DBC, mHdbc);
#else
        SQLFreeConnect(mHdbc);
#endif
        mHdbc = NULL;
    }

    if (mHenv)
    {
#if defined(HAVE_SQLFREEHANDLE) || defined(WIN32)
        SQLFreeHandle(SQL_HANDLE_ENV, mHenv);
#else
        SQLFreeEnv(mHenv);
#endif
        mHenv = NULL;
    }

    mIsConnected = false;
}

void DbConnection::EnumColumns(fieldlist_t& fields,
        const db_c_str_t tableName,
        const db_c_str_t colName,
        const db_c_str_t catalogName, 
        const db_c_str_t schemaName)
        throw (std::exception)
{
    DbCommand my_cmd(*this);

    if (!JGAADB_SQLOK(SQLColumns(my_cmd.GetHtsmt(),
        (SQLCHAR *)catalogName, catalogName ? SQL_NTS : 0,
        (SQLCHAR *)schemaName, schemaName ? SQL_NTS : 0,
        (SQLCHAR *)tableName, tableName ? SQL_NTS : 0,
        (SQLCHAR *)colName, colName ? SQL_NTS : 0)))
    {
        ThrowDb(this,
            SQL_HANDLE_DBC,
            GetHdbc(),
            "DbConnection::EnumColumns()",
            "SQLColumns() failed.",
            NULL);
    }

    // Fetch results
    DbRecordset results(&my_cmd);
    for(results.FetchFirst(); !results.IsEof(); results.MoveNext())
    {
        DbColumnDef col;
        col.mName = results[3].GetAsString();
        col.mSqlType = atoi(results[3].GetAsString());
        col.mIsNullable = atoi(results[10].GetAsString()) != SQL_NO_NULLS;
        col.mSize = atoi(results[6].GetAsString());
        col.mDigits = atoi(results[8].GetAsString());

        try
        {
            // ODBC v3
            col.mDefaultValue = results[12].GetAsString();
        }
        catch(...)
        {
        }

        fields.push_back(col);
    }
}

std::string DbConnection::GetValidName(const char *name)
{
	std::string rval = GetSqlQuoteChar();
	rval += name;
	rval += GetSqlQuoteChar();
	return rval;
}

//============================= CALLBACK   ===================================
//============================= ACCESS     ===================================

HENV DbConnection::GetHdbc() const
{
    return mHdbc;
}

HENV DbConnection::GetHenv() const
{
    return mHenv;
}

//============================= INQUIRY    ===================================

bool DbConnection::IsConnected() const
{
    return mIsConnected;
}

bool DbConnection::IsTable(const db_c_str_t tableName,
                           const db_c_str_t catalogName, 
                           const db_c_str_t schemaName)
                           throw (std::exception)
{
    db_table_list_t table_list;
    DbCommand my_cmd(*this);

    my_cmd.EnumTables(table_list);

    for(db_table_list_t::const_iterator P = table_list.begin()
        ; P != table_list.end()
        ; ++P)
    {
        if (catalogName && stricmp(P->GetCatalog().c_str(), catalogName))
            continue;
        
        if (schemaName && stricmp(P->GetSchema().c_str(), schemaName))
            continue;
        
        if (stricmp(P->GetName().c_str(), tableName) == 0)
            return true;
    }

    return false;
}

bool DbConnection::QueryInfo(SQLUSMALLINT infoType, 
							 std::string& infoBuffer)
{
#define BUF_LEN 1024 * 4
	SQLCHAR cbuf[BUF_LEN];
	char char_buf[BUF_LEN];
	SQLSMALLINT buf_len = sizeof(cbuf);
	bool rval = false;
	
	if (JGAADB_SQLOK(SQLGetInfo(mHdbc, 
		infoType, 
		(SQLPOINTER)cbuf, 
		buf_len, &buf_len)))
	{
		rval = true;
		SQLCHAR *p = cbuf;
		char *pp = char_buf;
		while(*p)
			*pp++ = *p++;
		*pp = 0;

		infoBuffer = char_buf;
	}

	return rval;
}

const std::string& DbConnection::GetDbmsName() const
{
	return mDbmsName;
}

const std::string& DbConnection::GetDbmsVersion() const
{
	return mDbmsVersion;
}

const std::string& DbConnection::GetSqlQuoteChar() const
{
	return mSqlQuoteChar;
}


/////////////////////////////// PROTECTED  ///////////////////////////////////

/////////////////////////////// PRIVATE    ///////////////////////////////////



END_JGAADB_NAMESPACE

