/* 
    jgaadb - is a simple C++ class library for ODBC databases.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 - 2002 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/

#include "StdAfx.h"
#include "jgaadb.h"

BEGIN_JGAADB_NAMESPACE

/////////////////////////////// PUBLIC ///////////////////////////////////////

//============================= LIFECYCLE ====================================

DbRecordset::DbRecordset(DbCommand *pCommand) :
mpCommand(pCommand),
mIsEof(false),mMustCloseCursor(false)
{
}// DbRecordset

DbRecordset::~DbRecordset()
{
    // Delete columns
    KillColumns();

}// ~DbRecordset


//============================= OPERATORS ====================================

DbColumn& DbRecordset::operator [] (unsigned index) throw (DbException)
{
    if (index >= mColumns.size())
        DbThrow("Column index out of range");

    return *mColumns[index];
}

//============================= OPERATIONS ===================================

void DbRecordset::Reset()
{
    KillColumns();
    mIsEof = false;
}

void DbRecordset::MoveNext() throw(std::exception)
{
    FetchNext();
}

//============================= CALLBACK   ===================================
//============================= ACCESS     ===================================
//============================= INQUIRY    ===================================

bool DbRecordset::IsEof() const
{
    return mIsEof;
}

size_t DbRecordset::GetNumColumns() const
{
    return DbRecordset::mColumns.size();
}

/////////////////////////////// PROTECTED  ///////////////////////////////////

void DbRecordset::KillColumns()
{
    // Delete all columns
    {
        for(columns_t::iterator P = mColumns.begin()
            ; P != mColumns.end()
            ; ++P)
        {
            DbColumn *ptr = *P;
            if (ptr)
            {
                delete ptr;
                *P = NULL;
            }
        }
    }

	if (mMustCloseCursor)
	{
		SQLFreeStmt(mpCommand->GetHtsmt(), SQL_CLOSE);
	}
}

void DbRecordset::FetchFirst() throw(std::exception)
{
    /* A SQL query has been performed. We will
       prepere the internal data structure, and
       return the first row (if present).
    */

    // Enumerate columns
    SQLSMALLINT num_cols = 0;
    if (!JGAADB_SQLOK(SQLNumResultCols(mpCommand->GetHtsmt(), &num_cols)))
    {
        ThrowDb(&mpCommand->GetConn(), 
            SQL_HANDLE_STMT, 
            mpCommand->GetHtsmt(),
            "DbRecordset::FetchFirst()",
            "SQLNumResultCols() failed",
            NULL);
    }

    mColumns.resize(num_cols, NULL);

    for(SQLSMALLINT index = 0; index < num_cols; index++)
    {
        SQLSMALLINT col_name_len = 0,
            sql_type = 0,
            col_scale = 0,
            col_nullable = 0;
        SQLUINTEGER col_size = 0;
        
        SQLCHAR col_name_buf[1024];

        if (!JGAADB_SQLOK(SQLDescribeCol(mpCommand->GetHtsmt(), 
            index +1,
            col_name_buf,
            sizeof(col_name_buf),
            &col_name_len,
            &sql_type,
            &col_size,
            &col_scale,
            &col_nullable)))
        {
            ThrowDb(&mpCommand->GetConn(), 
                SQL_HANDLE_STMT, 
                mpCommand->GetHtsmt(),
                "DbRecordset::FetchFirst()",
                "SQLDescribeCol() failed",
                NULL);
        }

        DbColumn *pnew_col = new DbColumn((const char *)col_name_buf,
            sql_type,
            col_size,
            col_scale,
            SQL_NULLABLE == col_nullable);
        
        mColumns[index] = pnew_col;
        
        if (!JGAADB_SQLOK(SQLBindCol(mpCommand->GetHtsmt(),
            index +1,
            pnew_col->GetCtype(),
            pnew_col->GetPointer(),
            pnew_col->GetBufferLength(),
            &pnew_col->mDataLenUsed)))
        {
            ThrowDb(&mpCommand->GetConn(), 
                SQL_HANDLE_STMT, 
                mpCommand->GetHtsmt(),
                "DbRecordset::FetchFirst()",
                "SQLBindCol() failed",
                NULL);
        }
    }

    FetchNext();
	mMustCloseCursor = true;
}

void DbRecordset::ResetCache(DbColumn *pCol)
{
    pCol->ResetCache();
}

void DbRecordset::FetchNext() throw(std::exception)
{
    SQLRETURN result = SQLFetch(mpCommand->GetHtsmt());
    //extern int debug_flag;

    std::for_each(mColumns.begin(), mColumns.end(), ResetCache);

    if (SQL_SUCCESS == result)
    {
        mIsEof = false;
    }
    else if (SQL_SUCCESS_WITH_INFO == result)
    {
        // See if we can handle the situation...
        for(int index = 1;; index++)
        {
            SQLCHAR sql_state[6];
            SQLINTEGER sql_native_error = 0;
            SQLCHAR sql_error_message[250];
            SQLSMALLINT sql_error_message_len = 0;

            result = SQLGetDiagRec(SQL_HANDLE_STMT,
                mpCommand->GetHtsmt(),
                index,
                (SQLCHAR *)sql_state,
                &sql_native_error,
                (SQLCHAR *)sql_error_message,
                sizeof(sql_error_message),
                &sql_error_message_len);

            if (SQL_NO_DATA == result)
                break;

            /*if (debug_flag)
            {
                std::cout 
                    << " - db - FetchNext(): SQL_SUCCESS_WITH_INFO sql state: "
                    << sql_state
                    << " Native error: "
                    << sql_native_error
                    << " Error message: "
                    << sql_error_message
                    << " index: "
                    << index
                    <<  std::endl;
            }*/

            if (!JGAADB_SQLOK(result))
            {
                ThrowDb(&mpCommand->GetConn(), 
                    SQL_HANDLE_STMT, 
                    mpCommand->GetHtsmt(),
                    "DbRecordset::FetchNext()",
                    "SQLGetDiagRec() failed",
                    NULL);
            }

            if ((sql_state[0] == '0')
                && (sql_state[1] == '1')
                && (sql_state[2] == 'S')
                && (sql_state[3] == '0')
                && (sql_state[4] == '7'))
            {
                // Fractional truncation. Ignore.
            }
            else if ((sql_state[0] == '0')
                && (sql_state[1] == '1')
                && (sql_state[2] == '0')
                && (sql_state[3] == '0')
                && (sql_state[4] == '4'))
            {
                // Get the field
                SQLINTEGER col_index = 0;
                SQLSMALLINT foo = 0;
                if (!JGAADB_SQLOK(SQLGetDiagField(SQL_HANDLE_STMT,
                    mpCommand->GetHtsmt(),
                    index,
                    SQL_DIAG_COLUMN_NUMBER,
                    &col_index,
                    sizeof(col_index),
                    &foo)))
                {
                    ThrowDb(&mpCommand->GetConn(), 
                        SQL_HANDLE_STMT, 
                        mpCommand->GetHtsmt(),
                        "DbRecordset::FetchNext()",
                        "SQLGetDiagField() failed",
                        NULL);
                }

                // Truncated buffer. 
                DbColumn *pcol = &operator [] (col_index -1);

                if (SQL_NO_TOTAL == pcol->mDataLenUsed)
                {
                    ThrowDb(&mpCommand->GetConn(), 
                        SQL_HANDLE_STMT, 
                        mpCommand->GetHtsmt(),
                        "DbRecordset::FetchNext()",
                        "SQL_NO_TOTAL in field length. Data buffer is probarbly too large",
                        NULL);
                }

                // Adjust the buffer size to the truncated size
                if (pcol->mDataLenUsed > pcol->mAllocatedBytes)
                    pcol->mDataLenUsed = pcol->mAllocatedBytes;
            }
            else
            {
                // Give up!
                ThrowDb(&mpCommand->GetConn(), 
                    SQL_HANDLE_STMT, 
                    mpCommand->GetHtsmt(),
                    "DbRecordset::FetchNext()",
                    "SQLFetch() == SQL_SUCCESS_WITH_INFO",
                    NULL);
            }
        }
    }
    else if (SQL_ERROR == result)
    {
         ThrowDb(&mpCommand->GetConn(), 
                SQL_HANDLE_STMT, 
                mpCommand->GetHtsmt(),
                "DbRecordset::FetchNext()",
                "SQLFetch() == SQL_ERROR",
                NULL);
    }
    else if (SQL_NO_DATA == result)
    {
        if (mIsEof)
        {
            ThrowDb(&mpCommand->GetConn(), 
                SQL_HANDLE_STMT, 
                mpCommand->GetHtsmt(),
                "DbRecordset::FetchNext()",
                "EOF is reached",
                NULL);
        }
        mIsEof = true;
    }
    else
    {
        char errsmg[128];
        sprintf(errsmg, "SQLFetch() returned unexcpeted error %d", result);
        
        ThrowDb(&mpCommand->GetConn(), 
            SQL_HANDLE_STMT, 
            mpCommand->GetHtsmt(),
            "DbRecordset::FetchNext()",
            errsmg,
            NULL);
    }
}

/////////////////////////////// PRIVATE    ///////////////////////////////////

END_JGAADB_NAMESPACE
